package jfd.ale;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.awt.Point;
import java.util.Hashtable;
import java.lang.reflect.Method;

class ConstraintLayout implements LayoutManager2, MethodTableReciever
{
   private static Hashtable nameToMethod    = new Hashtable();
   private Hashtable nameToComponent        = null;
   private Hashtable componentToConstraints = null;

   public static final String LEFT         = "l";
   public static final String RIGHT        = "r";
   public static final String TOP          = "t";
   public static final String BOTTOM       = "b";
   public static final String X            = "x";
   public static final String Y            = "y";
   public static final String WIDTH        = "w";
   public static final String HEIGHT       = "h";

   private static final String methodNames[] =
   {
      LEFT, RIGHT, TOP, BOTTOM, X, Y, WIDTH, HEIGHT
   };

   private Object enclosingObject        = null;
   private Hashtable externalMethodTable = null;

   private Dimension size                = null;

   static
   {
      Class parameterTypes[] = new Class[1];
      parameterTypes[0]      = Component.class;
      boolean error          = true;
      int i                  = 0;

      try
      {
         Class c = ConstraintLayout.class;

         while(i < methodNames.length)
         {
            nameToMethod.put(
               methodNames[i], 
               c.getMethod(methodNames[i], parameterTypes)
            );

            i++;
         }

         error = false;
      }
      catch(Exception e) {}

      if(error)
      {
         System.out.println(
            "ConstraintLayout, static initialization section\n" +
            "could not look up method: [" + methodNames[i] + "]"
         );

         System.exit(-1);
      }
   }

   public ConstraintLayout(Object enclosingObject)
   {
      nameToComponent        = new Hashtable();
      componentToConstraints = new Hashtable();
      this.enclosingObject   = enclosingObject;
   }

   // Fulfill our MethodTableReciever interface

   public void recieve(Hashtable methodTable)
   {
      externalMethodTable = methodTable;
   }

   public static boolean isLocationMethod(String name)
   {
      return(
         name.equals(LEFT)   ||
         name.equals(RIGHT)  ||
         name.equals(TOP)    ||
         name.equals(BOTTOM) ||
         name.equals(X)      ||
         name.equals(Y)
      );
   }

   public void addLayoutComponent(String s, Component c)
   {
      System.out.println(
         "ConstraintLayout.addLayoutComponent(String, Component)\n" +
         "incompatible method call, use addLayoutComponent(Component, Object)"
      );
      
      System.exit(-1);
   }

   public void addLayoutComponent(Component component, Object constraints)
   {
      Constraints theConstraints = (Constraints)constraints;

      String name = theConstraints.getName();

      nameToComponent.put(name,   component);
      componentToConstraints.put(component, theConstraints);
   }

   public void removeLayoutComponent(Component c)
   {
      Constraints constraints = (Constraints)componentToConstraints.get(c);

      if(constraints == null)
      {
         System.out.println(
            "ConstraintLayout.removeLayoutComponent(Component)\n" +
            "attempt to remove unknown Component: [" + c.toString() + "]"
         );

         System.exit(-1);
      }

      nameToComponent.remove(constraints.getName());
      componentToConstraints.remove(c);

      // Our container's size will likely need to be recalculated

      size = null;
   }

   private Dimension getSize(Container parent)
   {
      if(size == null)
         layoutContainer(parent);

       return(size);
   }

   public void invalidateLayout(Container parent) {}

   public float getLayoutAlignmentX(Container parent)
   {
      return(0.5f);
   }

   public float getLayoutAlignmentY(Container parent)
   {
      return(0.5f);
   }

   public Dimension maximumLayoutSize(Container parent)
   {
      return(preferredLayoutSize(parent));
   }

   public Dimension minimumLayoutSize(Container parent)
   {
      return(preferredLayoutSize(parent));
   }

   public Dimension preferredLayoutSize(Container parent)
   {
      return(getSize(parent));
   }

   public void layoutContainer(Container parent)
   {
      // Many of our components will likely refer to "parent"
      // so establish that reference now

      nameToComponent.put("parent", parent);

      Component array[] = parent.getComponents();

      // Size all Components first.  This allows a Component's
      // location to be based on the height or width of a
      // Component that is added to the Container later.

      int widths[]    = new int[array.length];
      int heights[]   = new int[array.length];

      for(int i = 0; i < array.length; i ++)
      {
         // Each component may also refer to itself via "this".
         // Ensure that for this component "this" is properly set

         nameToComponent.put("this", array[i]);

         // We'll likely need the width/height of each Component
         // in the section below where we establish the Component's
         // location so save them off here.

         Constraints c = (Constraints)componentToConstraints.get(array[i]);

         widths[i]  = (int)(c.getWidth().execute(this)  + 0.5);
         heights[i] = (int)(c.getHeight().execute(this) + 0.5);

         // We use 0 to indicate "use the preferred size" for height
         // and width.  Handle this here

         if(widths[i] == 0 || heights[i] == 0)
         {
            Dimension d = array[i].getPreferredSize();

            if(widths[i] == 0)
               widths[i] = d.width;

            if(heights[i] == 0)
               heights[i] = d.height;
         }

         array[i].setSize(widths[i], heights[i]);
      }

      int xMin = 0;
      int yMin = 0;
      int xMax = 0;
      int yMax = 0;

      // Now position all the Components

      for(int i = 0; i < array.length; i ++)
      {
         // Each component may also refer to itself via "this".
         // Ensure that for this component "this" is properly set

         nameToComponent.put("this", array[i]);

         Constraints c = (Constraints)componentToConstraints.get(array[i]);

         int x = (int)(c.getX().execute(this) + 0.5);
         int y = (int)(c.getY().execute(this) + 0.5);

         // Based upon alignment, work out the coordinates of the
         // component's upper left corner

         int left = x;
         int top  = y;

         Align align    = c.getAlign();
         String hString = align.getHorizAlign();
         String vString = align.getVertAlign();

         if(hString.equals(Align.CENTER))
         {
            left -= widths[i] / 2;
         }
         else if(hString.equals(Align.RIGHT))
         {
            left -= widths[i];
         }

         if(vString.equals(Align.CENTER))
         {
            top -= heights[i] / 2;
         }
         else if(vString.equals(Align.BOTTOM))
         {
            top -= heights[i];
         }

         // Record the bounds of all components in this container for
         // setting the container's size

         if(left < xMin)                  xMin = left;
         else if(left + widths[i] > xMax) xMax = left + widths[i];

         if(top < yMin)                   yMin = top;
         else if(top + heights[i] > yMax) yMax = top + heights[i];

         array[i].setLocation(left, top);
      }

      size = new Dimension((xMax - xMin) + 1, (yMax - yMin) + 1);
   }

   public double l(Component c)
   {
      Point p = c.getLocation();
      return((double)p.x);
   }

   public double t(Component c)
   {
      Point p = c.getLocation();
      return((double)p.y);
   }

   public double r(Component c)
   {
      Point p     = c.getLocation();
      Dimension d = c.getSize();
      return((double)(p.x + d.width));
   }

   public double b(Component c)
   {
      Point p     = c.getLocation();
      Dimension d = c.getSize();
      return((double)(p.y + d.height));
   }

   public double x(Component c)
   {
      Point p     = c.getLocation();
      Dimension d = c.getSize();
      return((double)(p.x + d.width / 2));
   }

   public double y(Component c)
   {
      Point p     = c.getLocation();
      Dimension d = c.getSize();
      return((double)(p.y + d.height / 2));
   }

   public double w(Component c)
   {
      Dimension d = c.getSize();
      return((double)d.width);
   }

   public double h(Component c)
   {
      Dimension d = c.getSize();
      return((double)d.height);
   }

   public double invoke(
      String name,
      String methodName
   )
   {
      double value  = 0.0;
      Component c   = (Component)nameToComponent.get(name);

      if(c == null)
      {
         System.out.println(
            "ConstraintLayout.invoke(String, String)\n" +
            "no such Component: [" + name + "]"
         );

         System.exit(-1);
      }
      else
      {
         Object args[] = new Object[1];
         args[0]       = c;
   
         try
         {
            Method m = (Method)nameToMethod.get(methodName);
            value    = ((Double)m.invoke(this, args)).doubleValue();
         }
         catch(Exception e)
         {
            System.out.println(
               "ConstraintLayout.invoke(String, String)\n" +
               "invocation failed for: [" + name + "." + methodName + "]"
            );

            System.exit(-1);
         }
      }
      return(value);
   }

   public double invokeExternal(String methodName)
   {
      // Immediate sanity check

      if(externalMethodTable == null)
      {
         System.out.println(
            "ConstraintLayout: internal error, Hashtable " + 
            "\"externalMethodTable\" is null"
         );

         System.exit(-1);
      }

      double d      = 0.0;
      boolean error = true;

      try
      {
         Method m = (Method)externalMethodTable.get(methodName);

         if(m != null)
         {
            Double dValue =
               (Double)m.invoke(enclosingObject, new Object[] {});

            d     = dValue.doubleValue();
            error = false;
         }
      }
      catch(Exception e) {}

      if(error)
      {
         // We shouldn't be here, "Section" should have verified
         // the existence/correctness of all ExternalMethodNodes already

         System.out.println(
            "Error invoking \"" + methodName + 
            "()\" on object \"" + enclosingObject + "\""
         );

         System.exit(-1);
      }

      return(d);
   }
}
